from ast import literal_eval
from pathlib import Path
from urllib.parse import urlsplit, urlunsplit

import aiosqlite
import bs4
import httpx
from platformdirs import user_data_dir
from textual import Logger


class CodeshareClient:
    def __init__(self, logger: Logger) -> None:
        self._data_dir = Path(user_data_dir("frics", "frics", ensure_exists=True))
        self._db_path = self._data_dir / "codeshare.db"
        self._log = logger

    def db_exists(self) -> bool:
        """Check if the database exists."""
        return self._db_path.exists()

    async def _ensure_db_schema(self) -> None:
        """Ensure the database schema is up to date."""
        async with aiosqlite.connect(self._db_path) as db:
            # We're using FTS5 for full-text search
            await db.execute("DROP TABLE IF EXISTS projects")
            await db.execute(
                """
                CREATE TABLE IF NOT EXISTS projects
                (
                    id TEXT PRIMARY KEY,
                    project_name TEXT,
                    author TEXT,
                    frida_version TEXT,
                    description TEXT,
                    source TEXT
                )
                """
            )

    async def update_db(self) -> None:
        """Update the database of codeshare snippets."""
        await self._ensure_db_schema()
        async with httpx.AsyncClient(base_url="https://codeshare.frida.re") as client:
            page = 1
            while True:
                response = await client.get("browse", params={"page": page}, timeout=10)

                if response.status_code != 200:
                    break

                # Parse the projects and extract the hrefs
                soup = bs4.BeautifulSoup(response.text, "html.parser")
                for article in soup.find_all("article"):
                    project_url = article.find("a").get("href")
                    split_url = urlsplit(project_url.replace("@", ""))
                    project_api_url: str = urlunsplit(split_url._replace(path="/api/project" + split_url.path))

                    # Get the project details
                    author = split_url.path.split("/")[1]
                    api_response = await client.get(project_api_url, timeout=10)
                    if api_response.status_code == 200:
                        project = api_response.json()

                        project_id = project["id"]
                        project_name = project["project_name"]
                        project_description = project["description"]
                        project_frida_version = project["frida_version"]
                        project_source = project["source"]

                        # Frida version from API is either a string or a list as a string
                        # We try to use literal eval to get the most recent version if a list is provided
                        try:
                            frida_version = literal_eval(project["frida_version"])
                            project_frida_version = frida_version[0]
                        except Exception:
                            pass

                        async with aiosqlite.connect(self._db_path) as db:
                            await db.execute(
                                "INSERT OR REPLACE INTO projects VALUES (?, ?, ?, ?, ?, ?)",
                                (
                                    project_id,
                                    project_name,
                                    author,
                                    project_frida_version,
                                    project_description,
                                    project_source,
                                ),
                            )
                            await db.commit()
                page += 1

    async def query_all(self):
        """Query all projects."""

        async with aiosqlite.connect(self._db_path) as db:
            async with db.execute("""SELECT id, project_name, author, frida_version FROM projects""") as cursor:
                async for row in cursor:
                    yield list(row)

    async def get_project(self, project_id: str) -> list[str] | None:
        """Query a project by ID."""

        async with aiosqlite.connect(self._db_path) as db:
            async with db.execute(
                """SELECT description, source FROM projects WHERE id=?""",
                (project_id,),
            ) as cursor:
                async for row in cursor:
                    if row:
                        return list(row)
                    else:
                        raise ValueError(f"Project with ID {project_id} not found")

    async def search(self, search_term: str):
        """Search for a project by term."""

        like_search_term = f"%{search_term}%"
        async with aiosqlite.connect(self._db_path) as db:
            try:
                # We need to search all columns for the search term
                # We use LIKE to do a case-insensitive search
                # FTS5 doesn't support LIKE, so we can't use it :(
                async with db.execute(
                    """
                    SELECT id, project_name, author, frida_version
                    FROM projects
                    WHERE project_name LIKE ?
                    OR author LIKE ?
                    OR description LIKE ?
                    OR source LIKE ?
                    """,
                    (
                        like_search_term,
                        like_search_term,
                        like_search_term,
                        like_search_term,
                    ),
                ) as cursor:
                    async for row in cursor:
                        yield list(row)
            except aiosqlite.OperationalError:
                return



import re

import pyperclip
from rich.syntax import Syntax
from textual import on, work
from textual.app import App, ComposeResult
from textual.containers import Horizontal, Vertical
from textual.widgets import (
    Button,
    DataTable,
    Header,
    Input,
    Label,
    LoadingIndicator,
    TextLog,
)

from frics.codeshare import CodeshareClient


class Frics(App):
    CSS_PATH = "frics.css"
    TITLE = "FЯICS"
    SUB_TITLE = "Frida Codeshare Explorer"

    def compose(self) -> ComposeResult:
        self.dark = True
        self._codeshare_client = CodeshareClient(self.log)

        yield Header()
        with Horizontal(name="root"):
            with Vertical(name="left", id="left_vert"):
                yield Input(placeholder="Search...", name="input_search", id="input_search")
                with Horizontal(name="update_container", id="update_container"):
                    yield Button(label="Update DB", name="btn_update_db", id="btn_update_db")
                    yield LoadingIndicator(name="loading", id="loading")
                yield DataTable(name="dt_db", id="dt_db")
                yield Label("Description", name="lbl_desc", id="lbl_desc")
                yield TextLog(name="text_desc", id="text_desc", wrap=True, highlight=True)
            with Vertical(name="right", id="right_vert"):
                yield Button(label="Copy to clipboard", name="btn_copy", id="btn_copy")
                yield TextLog(name="text_source", id="text_source", wrap=True, highlight=True)

    async def start_loading(self):
        """Start the loading indicator."""
        loading = self.query_one("#loading", LoadingIndicator)
        loading.styles.visibility = "visible"

    async def stop_loading(self):
        """Stop the loading indicator."""
        loading = self.query_one("#loading", LoadingIndicator)
        loading.styles.visibility = "hidden"

    @work(exclusive=True)
    @on(Button.Pressed, "#btn_update_db")
    async def update_db(self, message: Button.Pressed):
        """Update the database of codeshare snippets.

        Args:
            message (Button.Pressed): Button pressed message.
        """
        self.log("Updating database...")
        self.notify("Updating FЯICS database...")
        await self.start_loading()
        message.button.disabled = True

        await self._codeshare_client.update_db()
        self.notify("FЯICS database updated!")
        message.button.disabled = False
        await self.stop_loading()

        db_table = self.query_one(DataTable)
        db_table.clear()
        async for row in self._codeshare_client.query_all():
            db_table.add_row(*row)

    @on(Button.Pressed, "#btn_copy")
    def copy_to_clipboard(self, message: Button.Pressed):
        """Copy source code to the clipboard.

        Args:
            message (Button.Pressed): Button pressed message.
        """
        source_box = self.query_one("#text_source", TextLog)
        copy_string = ""

        # We need to copy the text without the line numbers
        for line in source_box.lines:
            code_line, *_ = line.text.rpartition(" ")
            code_line = code_line.partition("  ")[-1]

            code_match = re.match(r"\s*\d+\s(.*)", code_line)
            self.log(code_match)
            if code := code_match.group(1):  # type: ignore
                copy_string += code + "\n"
            else:
                copy_string += "\n"
        pyperclip.copy(copy_string)
        self.notify("Source code copied to clipboard!")

    @work(exclusive=True)
    @on(Input.Changed, "#input_search")
    async def search(self, message: Input.Changed):
        """Search the database.

        Args:
            message (Input.Changed): Input changed message.
        """
        search_term = message.input.value
        if not search_term:
            return

        db_table = self.query_one("#dt_db", DataTable)
        db_table.clear()

        async for row in self._codeshare_client.search(search_term):
            db_table.add_row(*row)

    @work(exclusive=True)
    @on(DataTable.RowSelected, "#dt_db")
    async def get_row_data(self, message: DataTable.RowSelected):
        """Get the row data.

        Args:
            message (DataTable.RowSelected): Row selected message.
        """
        project_id: str = message.data_table.get_row(message.row_key)[0]
        try:
            description, source = await self._codeshare_client.get_project(project_id)  # type: ignore
            desc_box = self.query_one("#text_desc", TextLog)
            source_box = self.query_one("#text_source", TextLog)

            desc_box.clear()
            desc_box.write(description)

            source_box.clear()
            source_box.write(
                Syntax(
                    source,
                    "js",
                    line_numbers=True,
                    word_wrap=True,
                    background_color="#090100",
                    theme="github-dark",
                ),
                scroll_end=False,
                expand=True,
                shrink=False,
            )

        except Exception as ex:
            self.log(f"Error getting project: {ex}")
            self.notify(f"Error getting project: {ex}")
            return

    async def on_mount(self):
        """Mount the app."""
        db_table = self.query_one("#dt_db", DataTable)
        db_table.cursor_type = "row"
        db_table.add_column("ID", width=0, key="project_id")
        db_table.add_column("Project Name", width=50, key="project_name")
        db_table.add_column("Author", width=15, key="author")
        db_table.add_column("Frida Version", key="frida_version")

        if not self._codeshare_client.db_exists():
            self.notify("FЯICS database not found. Please update the database.")
        else:
            async for row in self._codeshare_client.query_all():
                db_table.add_row(*row)


def run_app():
    """Run the app."""
    app = Frics()
    app.run()


if __name__ == "__main__":
    run_app()
